@@ -1,24 +1,3 @@
-/*
-  Racket
-  Copyright (c) 2004-2017 PLT Design Inc.
-
-  This library is free software; you can redistribute it and/or
-  modify it under the terms of the GNU Library General Public
-  License as published by the Free Software Foundation; either
-  version 2 of the License, or (at your option) any later version.
-
-  This library is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-  Library General Public License for more details.
-
-  You should have received a copy of the GNU Library General Public
-  License along with this library; if not, write to the Free
-  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-  Boston, MA 02110-1301 USA.
-
-*/
-
 #include "schpriv.h"
 #include "schmach.h"
 
@@ -161,7 +140,7 @@ typedef struct Scheme_Deferred_Expr {
      only need to process it once */
   int done;
 
-  /* the expression that has been deferred -- usally a Scheme_Lambda,
+  /* the expression that has been deferred -- usually a Scheme_Lambda,
      but can be a Scheme_IR_Local for propagation */
   Scheme_Object *expr;
 
@@ -333,15 +312,15 @@ static void letrec_check_lets_resume(Letrec_Check_Frame *frame, Scheme_IR_Let_He
        that we have invalidated; i.e., adding check-undefineds around
        references means there is one (more) instance where the LHS
        variable is not used in application position */
-    k = head->count;
+    k = 0;
     for (i = head->num_clauses; i--;) {
       irlv = (Scheme_IR_Let_Value *) body;
-      k -= irlv->count;
       for (j = 0; j < irlv->count; j++) {
         was_checked = (frame->ref[k + j] & LET_CHECKED);
         if (was_checked)
           irlv->vars[j]->non_app_count = irlv->vars[j]->use_count;
       }
+      k += irlv->count;
       body = irlv->body;
     }
   }
@@ -392,6 +371,13 @@ static Scheme_Object *letrec_check_local(Scheme_Object *o, Letrec_Check_Frame *f
         ls = (Scheme_Object *)make_deferred_expr_closure(o, frame);
         ls = scheme_make_pair(ls, outer_frame->def[dpos]);
         outer_frame->def[dpos] = ls;
+
+        /* But if we're trying to propagate to a variable that was
+           already referenced, we need to treat this one as
+           referenced, too */
+        if ((outer_frame->ref[dpos] & LET_APPLY_USE)
+            && in_frame->ref)
+          in_frame->ref[in_position] |= LET_APPLY_USE;
       }
     }
   }
@@ -425,10 +411,9 @@ static Scheme_Object *letrec_check_local(Scheme_Object *o, Letrec_Check_Frame *f
   return o;
 }
 
-static int is_effect_free_prim(Scheme_Object *rator)
+static int is_effect_free_prim(Scheme_Object *rator, int argc)
 {
-  if (SCHEME_PRIMP(rator)
-      && (SCHEME_PRIM_PROC_OPT_FLAGS(rator) & SCHEME_PRIM_IS_OMITABLE_ANY))
+  if (SCHEME_PRIMP(rator) && scheme_is_omitable_primitive(rator, argc))
     return 1;
 
   return 0;
@@ -445,7 +430,7 @@ static Scheme_Object *letrec_check_application(Scheme_Object *o, Letrec_Check_Fr
   /* we'll have to check the rator and all the arguments */
   n = 1 + app->num_args;
 
-  if (is_effect_free_prim(app->args[0])) {
+  if (is_effect_free_prim(app->args[0], app->num_args)) {
     /* an immediate prim cannot call anything among its arguments */
   } else {
     /* argument might get applied */
@@ -467,7 +452,7 @@ static Scheme_Object *letrec_check_application2(Scheme_Object *o, Letrec_Check_F
 
   app = (Scheme_App2_Rec *)o;
     
-  if (is_effect_free_prim(app->rator)) {
+  if (is_effect_free_prim(app->rator, 1)) {
     /* an immediate prim cannot call anything among its arguments */
   } else {
     /* argument might get applied */
@@ -489,7 +474,7 @@ static Scheme_Object *letrec_check_application3(Scheme_Object *o, Letrec_Check_F
 
   app = (Scheme_App3_Rec *)o;
 
-  if (is_effect_free_prim(app->rator)) {
+  if (is_effect_free_prim(app->rator, 2)) {
     /* an immediate prim cannot call anything among its arguments */
   } else {
     /* argument might get applied */
@@ -814,14 +799,11 @@ static Scheme_Object *letrec_check_define_values(Scheme_Object *lam, Letrec_Chec
   if (SCHEME_VEC_SIZE(lam) <= 1)
     return lam;
   else {
-    Scheme_Object *vars = SCHEME_VEC_ELS(lam)[0];
-    Scheme_Object *val = SCHEME_VEC_ELS(lam)[1];
-    SCHEME_ASSERT(SCHEME_PAIRP(vars) || SCHEME_NULLP(vars),
-                  "letrec_check_define_values: processing resolved code");
+    Scheme_Object *val = SCHEME_VEC_ELS(lam)[0];
 
     val = letrec_check_expr(val, frame, pos);
 
-    SCHEME_VEC_ELS(lam)[1] = val;
+    SCHEME_VEC_ELS(lam)[0] = val;
   }
     
   return lam;
@@ -860,55 +842,68 @@ static Scheme_Object *letrec_check_set(Scheme_Object *o, Letrec_Check_Frame *fra
       Scheme_App3_Rec *app3;
       Scheme_Object *name;
       Scheme_Sequence *seq;
-      
+      Scheme_IR_Let_Header *head;
+      Scheme_IR_Let_Value *lv;
+      Scheme_IR_Local *var, **vars;
+
+      /* Change
+           (set! <id> <rhs>)
+
+         to
+
+           (let ([<tmp> <rhs>])
+             (begin
+               <check>
+               (set! <id> <tmp>)))
+      */
+
       name = record_checked((Scheme_IR_Local *)sb->var, frame);
-      
+
       app3 = MALLOC_ONE_TAGGED(Scheme_App3_Rec);
       app3->iso.so.type = scheme_application3_type;
       app3->rator = scheme_check_assign_not_undefined_proc;
       app3->rand1 = sb->var;
       app3->rand2 = name;
-      
+
       seq = scheme_malloc_sequence(2);
       seq->so.type = scheme_sequence_type;
       seq->count = 2;
       seq->array[0] = (Scheme_Object *)app3;
       seq->array[1] = (Scheme_Object *)sb;
 
-      return (Scheme_Object *)seq;
+      if (SCHEME_TYPE(sb->val) > _scheme_ir_values_types_) {
+        /* obviously simple enough to not need a `let` wrapper */
+        return (Scheme_Object *)seq;
+      } else {
+        var = scheme_make_ir_local(scheme_intern_symbol("tmp"));
+        vars = MALLOC_N(Scheme_IR_Local*, 1);
+        vars[0] = var;
+        var->use_count = 1;
+        var->non_app_count = 1;
+
+        lv = MALLOC_ONE_TAGGED(Scheme_IR_Let_Value);
+        lv->iso.so.type = scheme_ir_let_value_type;
+        lv->count = 1;
+        lv->value = sb->val;
+        lv->body = (Scheme_Object *)seq;
+        lv->vars = vars;
+
+        head = MALLOC_ONE_TAGGED(Scheme_IR_Let_Header);
+        head->iso.so.type = scheme_ir_let_header_type;
+        head->count = 1;
+        head->num_clauses = 1;
+        head->body = (Scheme_Object *)lv;
+
+        sb->val = (Scheme_Object *)var;
+
+        return (Scheme_Object *)head;
+      }
     }
   }
 
   return o;
 }
 
-static Scheme_Object *letrec_check_define_syntaxes(Scheme_Object *lam, Letrec_Check_Frame *frame, Scheme_Object *pos)
-{
-  Scheme_Object *val;
-  val = SCHEME_VEC_ELS(lam)[3];
-
-  val = letrec_check_expr(val, frame, pos);
-  SCHEME_VEC_ELS(lam)[3] = val;
-
-  return lam;
-}
-
-static Scheme_Object *letrec_check_begin_for_syntax(Scheme_Object *lam, Letrec_Check_Frame *frame, Scheme_Object *pos)
-{
-  Scheme_Object *l, *a, *val;
-    
-  l = SCHEME_VEC_ELS(lam)[2];
-    
-  while (!SCHEME_NULLP(l)) {
-    a = SCHEME_CAR(l);
-    val = letrec_check_expr(a, frame, pos);
-    SCHEME_CAR(l) = val;
-    l = SCHEME_CDR(l);
-  }
-    
-  return lam;
-}
-
 static Scheme_Object *letrec_check_case_lambda(Scheme_Object *o, Letrec_Check_Frame *frame, Scheme_Object *pos)
 {
   Scheme_Case_Lambda *cl;
@@ -959,48 +954,11 @@ static Scheme_Object *letrec_check_apply_values(Scheme_Object *lam, Letrec_Check
   return lam;
 }
 
-static Scheme_Object *letrec_check_module(Scheme_Object *o, Letrec_Check_Frame *frame, Scheme_Object *pos)
-{
-  int i, cnt;
-  Scheme_Module *m;
-  Scheme_Object *val;
-  m = (Scheme_Module *)o;
-
-  if (!m->comp_prefix) {
-    /* already resolved */
-    return (Scheme_Object *)m;
-  }
-    
-  cnt = SCHEME_VEC_SIZE(m->bodies[0]);
-  for(i = 0; i < cnt; i++) {
-    val = SCHEME_VEC_ELS(m->bodies[0])[i];
-    val = letrec_check_expr(val, frame, pos);
-    SCHEME_VEC_ELS(m->bodies[0])[i] = val;
-  }
-
-  {
-    /* check submodules */
-    int k;
-    Scheme_Object *p;
-    for (k = 0; k < 2; k++) {
-      p = (k ? m->post_submodules : m->pre_submodules);
-      if (p) {
-        while (!SCHEME_NULLP(p)) {
-          letrec_check_expr(SCHEME_CAR(p), frame, pos);
-          p = SCHEME_CDR(p);
-        }
-      }
-    }
-  }
-
-  return o;
-}
-
 static Scheme_Object *letrec_check_k(void)
 {
   Scheme_Thread *p = scheme_current_thread;
   Scheme_Object *expr = (Scheme_Object *)p->ku.k.p1;
-  Letrec_Check_Frame *frame = (Letrec_Check_Frame *)p->ku.k.p3;
+  Letrec_Check_Frame *frame = (Letrec_Check_Frame *)p->ku.k.p2;
   Scheme_Object *pos = (Scheme_Object *)p->ku.k.p3;
 
   p->ku.k.p1 = NULL;
@@ -1041,7 +999,6 @@ static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame
   case scheme_application3_type:
     return letrec_check_application3(expr, frame, pos);
   case scheme_sequence_type:
-  case scheme_splice_sequence_type:
     return letrec_check_sequence(expr, frame, pos);
   case scheme_branch_type:
     return letrec_check_branch(expr, frame, pos);
@@ -1053,10 +1010,7 @@ static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame
     return letrec_check_lets(expr, frame, pos);
   case scheme_ir_toplevel_type: /* var ref to a top level */
     return expr;
-  case scheme_ir_quote_syntax_type:
-    return expr;
   case scheme_variable_type:
-  case scheme_module_variable_type:
     scheme_signal_error("got top-level in wrong place");
     return 0;
   case scheme_define_values_type:
@@ -1065,10 +1019,6 @@ static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame
     return letrec_check_ref(expr, frame, pos);
   case scheme_set_bang_type:
     return letrec_check_set(expr, frame, pos);
-  case scheme_define_syntaxes_type:
-    return letrec_check_define_syntaxes(expr, frame, pos);
-  case scheme_begin_for_syntax_type:
-    return letrec_check_begin_for_syntax(expr, frame, pos);
   case scheme_case_lambda_sequence_type:
     return letrec_check_case_lambda(expr, frame, pos);
   case scheme_begin0_sequence_type:
@@ -1078,17 +1028,14 @@ static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame
   case scheme_with_immed_mark_type:
     scheme_signal_error("internal error: with-immediate-mark not expected before optimization");
     return NULL;
-  case scheme_require_form_type:
-    return expr;
-  case scheme_module_type:
-    return letrec_check_module(expr, frame, pos);
   default:
     return expr;
   }
 }
 
-Scheme_Object *scheme_letrec_check_expr(Scheme_Object *expr)
+Scheme_Linklet *scheme_letrec_check_linklet(Scheme_Linklet *linklet)
 {
+  int i, cnt;
   Scheme_Object *val;
   Scheme_Object *init_pos = scheme_false;
   Letrec_Check_Frame *frame;
@@ -1105,11 +1052,16 @@ Scheme_Object *scheme_letrec_check_expr(Scheme_Object *expr)
      positions. We use a list of numbers for the RHS of a
      `let[rec]-values` form with multiple variables. */
 
-  val = letrec_check_expr(expr, frame, init_pos);
+  cnt = SCHEME_VEC_SIZE(linklet->bodies);
+  for(i = 0; i < cnt; i++) {
+    val = SCHEME_VEC_ELS(linklet->bodies)[i];
+    val = letrec_check_expr(val, frame, init_pos);
+    SCHEME_VEC_ELS(linklet->bodies)[i] = val;
+  }
 
   clean_dead_deferred_expr(*frame->deferred_chain);
 
-  return val;
+  return linklet;
 }
 
 /*========================================================================*/
